/* Copyright 2007, 2008 DJ Delorie <dj@delorie.com>
   Distributed under the terms of the GNU General Public License
   version 3, see COPYING-3.0 for details.
*/
#include "pins.h"
#include "funcs.h"

#include "TCPIP Stack/TCPIP.h"
#include "globals.h"

#define SNOOZE_MINUTES 10

#define NO_SNOOZE 0xFF
static byte snooze_minute = NO_SNOOZE;
static byte alarm_end_minute = NO_SNOOZE;
static unsigned int alarm_mp3 = 0;
static unsigned int alarm_img = 0;
static char hostname[80];
static int host_len, host_port_len;
static unsigned int url_port;
static unsigned int url_ptr;
static unsigned int url_len;
static int timer;
static NODE_INFO host_ip;
static TCP_SOCKET *socket;

enum
{
  A_none,
  A_dns,
  A_arp,
  A_arp2,
  A_connect,
  A_connected,
  A_sendurl,
  A_sendhost,
  A_readhttp,
  A_readstatus,
  A_readheader,
  A_readnl2,
} alarm_state = A_none;

#define MP3() (socket == &mp3_url_socket)
#define GET() if (checksocket(&c)) break

static int
checksocket (byte * c)
{
  if (!TCPIsConnected (*socket))
    {
      if (MP3 ())
	{
	  cprintf ("%s:%d: socket closed during headers\n", __FILE__,
		   __LINE__);
	  mp3_play_beep ();
	}
      alarm_state = A_none;
      TCPDisconnect (*socket);
      *socket = INVALID_SOCKET;
      return 1;
    }
  if (!TCPIsGetReady (*socket))
    return 1;
  TCPGet (*socket, c);
  return 0;
}

void
alarm_task ()
{
  byte c;

  switch (alarm_state)
    {
    case A_none:
      break;

    case A_dns:
      if (!DNSIsResolved (&host_ip.IPAddr))
	break;
      if (!DNSEndUsage ())
	{
	  cprintf ("cannot resolve %s\n", hostname);
	  if (MP3 ())
	    mp3_play_beep ();
	  alarm_state = A_none;
	  break;
	}
      alarm_state++;
      //cprintf ("IP: %P\n", &host_ip.IPAddr);
      break;

    case A_arp:
      ARPResolve (&host_ip.IPAddr);
      timer = iTick ();
      alarm_state++;
      break;

    case A_arp2:
      if (!ARPIsResolved (&host_ip.IPAddr, &host_ip.MACAddr))
	{
	  if (iTick () - timer > 1 * TICK_SECOND)
	    alarm_state--;
	  break;
	}
      //cprintf ("mac: %x\n", &host_ip.MACAddr, sizeof (host_ip.MACAddr));
      alarm_state++;
      break;

    case A_connect:
      if (*socket != INVALID_SOCKET)
	TCPDisconnect (*socket);
      *socket = TCPConnect (&host_ip, url_port);
      if (*socket == INVALID_SOCKET)
	{
	  cprintf ("%s:%d: no tcp sockets\n", __FILE__, __LINE__);
	  alarm_state = A_none;
	  mp3_play_beep ();
	}
      timer = iTick ();
      alarm_state++;
      break;

    case A_connected:
      if (!TCPIsConnected (*socket))
	{
	  if (iTick () - timer > 5 * TICK_SECOND)
	    {
	      if (MP3 ())
		{
		  cprintf ("%s:%d: connection timed out %d %d\n", __FILE__,
			   __LINE__, iTick (), timer);
		  mp3_play_beep ();
		}
	      alarm_state--;
	      TCPDisconnect (*socket);
	      *socket = INVALID_SOCKET;
	    }
	  break;
	}
      if (!TCPIsPutReady (*socket))
	break;

      TCPPutROMString (*socket, (ROM BYTE *) "GET ");
      alarm_state++;
      break;

    case A_sendurl:
      cprintf("url: ");
      if (!TCPIsPutReady (*socket))
	{
	  TCPFlush (*socket);
	  break;
	}
      while (TCPIsPutReady (*socket) && url_len > 0)
	{
	  char c = eeprom_readc (url_ptr);
	  cprintf("%c", c);
	  TCPPut (*socket, c);
	  url_ptr++;
	  url_len--;
	}
      cprintf("\n");
      if (url_len)
	TCPFlush (*socket);
      else
	alarm_state++;
      break;

    case A_sendhost:
      TCPPutROMString (*socket, (ROM BYTE *) " HTTP/1.1\r\nHost: ");
      hostname[host_len] = ':';
      TCPPutString (*socket, (BYTE *) hostname);
      TCPPutROMString (*socket, (ROM BYTE *) "\r\n\r\n");
      alarm_state++;
      break;

    case A_readhttp:
      GET ();
      if (c == ' ')
	alarm_state++;
      break;

    case A_readstatus:
      GET ();
      if (c != ' ')
	{
	  //cprintf ("status: %c\n", c);
	  alarm_state++;
	}
      break;

    case A_readheader:
      GET ();
      if (c == '\n')
	alarm_state++;
      break;

    case A_readnl2:
      GET ();
      if (c == '\r')
	break;
      if (c != '\n')
	alarm_state--;
      alarm_state++;
      //cprintf ("STREAM\n");
      if (MP3 ())
	{
	  alarm_state = A_none;
	  mp3_play_url ();
	}
      break;

    }
}

void
alarm_check ()
{
  uint addr, size;
  AlarmStruct entry;
  byte weekmask = 0x40 >> dow;

  addr = get_toc (TOC_alarm, &size);

  while (size > sizeof (entry))
    {
      eeprom_read (addr, &entry, sizeof (entry));
      addr += sizeof (entry);
      size -= sizeof (entry);
      // Check the least likely match first.


      // time
      if (entry.minute != minute)
	continue;
      if (entry.hour != hour)
	continue;

      // overall enabler
      if (!(entry.dow & 0x80))
	continue;

      if (entry.month)
	{
	  if (entry.day != day)
	    continue;
	  if (entry.month != month)
	    continue;
	}
      else
	{
	  // weekday, date
	  cprintf("alarm dow: %b %b\n", entry.dow, weekmask);
	  if (!(entry.dow & weekmask))
	    continue;
	}

#if 1
  cprintf ("\033[32m%b:%b %b %b/%b\033[0m\n",
	   hour, minute, weekmask, month, day);
#endif
#if 1
  //      if (entry.dow)
	cprintf ("%b:%b %b %b/%b %w\n",
		 entry.hour, entry.minute, entry.dow, entry.month, entry.day,
		 entry.web);
#endif
      cprintf ("trigger!\n");
      trigger_alarm (entry.web, entry.zen);
      return;
    }

  if (minute == alarm_end_minute)
    alarm_off ();

  if (minute == snooze_minute)
    {
      cprintf ("snooze expired.\n");
      snooze_minute = NO_SNOOZE;
      alarm_reconnect ();
      return;
    }
}

// Returns nonzero if the fetch fails.
static int
fetch_url (unsigned int addr, TCP_SOCKET * sockptr)
{
  int hostp = 0;
  byte c;

  if (*sockptr != INVALID_SOCKET && TCPIsConnected (*sockptr))
    {
      TCPDisconnect (*sockptr);
      *sockptr = INVALID_SOCKET;
    }

  eeprom_read (addr, &url_len, 2);
  addr += 2;
  eeprom_read (addr, hostname, 7);
  hostname[7] = 0;
  cprintf ("i%w<%s>\n", addr, hostname);
  if (memcmp (hostname, "http://", 7))
    return 1;
  addr += 7;
  url_len -= 7;
  hostp = 0;
  while (1)
    {
      c = eeprom_readc (addr);
      if (c == ':' || c == '/' || url_len == 0)
	break;
      url_len--;
      addr++;
      hostname[hostp++] = c;
    }
  host_len = hostp;
  hostname[hostp++] = 0;
  if (c == ':')
    {
      addr++;
      url_len--;
      url_port = 0;
      while (1)
	{
	  c = eeprom_readc (addr);
	  if (c == '/' || url_len == 0)
	    break;
	  addr++;
	  url_len--;
	  url_port = url_port * 10;
	  url_port += c - '0';
	  hostname[hostp++] = c;
	}
      hostname[hostp++] = '\0';
    }
  else
    {
      hostname[hostp++] = '8';
      hostname[hostp++] = '0';
      hostname[hostp++] = '\0';
      url_port = 80;
    }
  host_port_len = hostp - 1;

  cprintf ("H<%s> p<%d>\n", hostname, url_port);

  DNSEndUsage ();
  DNSBeginUsage ();
  DNSResolve ((BYTE *) hostname, DNS_TYPE_A);

  socket = sockptr;
  url_ptr = addr;

  return 0;
}

void
trigger_alarm (unsigned int address, int zen)
{
  eeprom_read (address, &alarm_mp3, 2);
  eeprom_read (address, &alarm_img, 2);

  // We'll hit this again in an hour.
  alarm_end_minute = minute;
  cprintf ("alarm_end_minute now %b\n", alarm_end_minute);

  mp3_zen_timer_start (zen);
  alarm_reconnect ();
}

void
alarm_reconnect ()
{
  if (alarm_mp3 && !fetch_url (alarm_mp3, &mp3_url_socket))
    alarm_state = A_dns;
  else
    {
      cprintf ("%s:%d: no mp3 url\n", __FILE__, __LINE__);
      mp3_play_beep ();
      if (alarm_img)		// && !fetch_url (alarm_img, &img_url_socket))
	{
	  alarm_state = A_dns;
	}
    }
}

void
trigger_alarm_i (int index)
{
  byte zen;
  unsigned int addr;

  addr = get_toc (TOC_config, 0);
  zen = eeprom_readc (addr + ConfigOffset (zen_setting));
  addr = get_toc (TOC_web, 0);
  trigger_alarm (addr + index * sizeof (WebDestination), zen);
}

void
alarm_off ()
{
  mp3_stop ();
  alarm_state = A_none;
  snooze_minute = NO_SNOOZE;
  alarm_end_minute = NO_SNOOZE;
  sleep_time = 0;
}

void
alarm_snooze ()
{
  if (play_source == PLAY_NONE)
    return;
  mp3_stop ();
  alarm_state = A_none;
  sleep_time = 0;
  snooze_minute = bcd ((unbcd (minute) + SNOOZE_MINUTES) % 60);
  cprintf ("snooze set to %b\n", snooze_minute);
}

void
sleep_on ()
{
  uint addr;
  byte sleep_volume;

  addr = get_toc (TOC_config, 0);
  sleep_volume = eeprom_readc (addr + ConfigOffset (sleep_volume));
  trigger_alarm_i (0);
  mp3_zen_volume (sleep_volume);
}

void
sleep_off ()
{
  alarm_off ();
  mp3_zen_volume (255);
}
