/* Copyright 2007, 2008 DJ Delorie <dj@delorie.com>
   Distributed under the terms of the GNU General Public License
   version 3, see COPYING-3.0 for details.
*/
#include "pins.h"
#include "funcs.h"
#include "globals.h"
#include "TCPIP Stack/TCPIP.h"

static unsigned int last = 0xffff;

static TCP_SOCKET eeprom_server_socket;

#define WAIT waitms(1)

void
write_1 (unsigned int address, void *ptr, int len)
{
  while (i2c_wra (EEPROM_I2C, address, ptr, len))
    WAIT;
  last = address + len;
}

void
eeprom_write (unsigned int address, void *ptr, int len)
{
  unsigned int la = address + len - 1;
  while ((address ^ la) & 0xff80)
    {
      int l = 128 - (address & 0x7f);
      write_1 (address, ptr, l);
      address += l;
      ptr += l;
      len -= l;
    }
  write_1 (address, ptr, len);
}

void
eeprom_read (unsigned int address, void *ptr, int len)
{
  if (last != address)
    while (i2c_wra (EEPROM_I2C, address, 0, 0))
      WAIT;
  while (i2c_rd (EEPROM_I2C, ptr, len))
    WAIT;
  last = address + len;
}

int
eeprom_readc (unsigned int address)
{
  byte b;
  if (last != address)
    while (i2c_wra (EEPROM_I2C, address, 0, 0))
      WAIT;
  while (i2c_rd (EEPROM_I2C, &b, 1))
    WAIT;
  last = address + 1;
  return b;
}

static enum
{
  // Waiting for 'R' or 'W' token
  EE_IDLE,
  // Got 'R', waiting for addr/count
  EE_READADDR,
  // Sending data bytes
  EE_READING,
  // Got 'W', waiting for addr/count
  EE_WRITEADDR,
  // Reading data bytes
  EE_WRITING,
} eestate;

void
eeprom_init ()
{
  eeprom_server_socket = TCPListen (EEPROM_SERVER_PORT);
  eestate = EE_IDLE;
}

void
eeprom_task ()
{
  static unsigned int addr = 0, count = 0;

  byte c, ah, al;
  int space;
  byte buf[128];

  if (TCPIsConnected (eeprom_server_socket))
    {
      switch (eestate)
	{
	case EE_IDLE:
	  while (TCPGet (eeprom_server_socket, &c))
	    {
	      if (c == 'R')
		{
		  eestate = EE_READADDR;
		  break;
		}
	      if (c == 'W')
		{
		  eestate = EE_WRITEADDR;
		  break;
		}
	    }
	  break;

	case EE_READADDR:
	case EE_WRITEADDR:
	  if (TCPIsGetReady (eeprom_server_socket) < 3)
	    return;
	  TCPGet (eeprom_server_socket, &al);
	  TCPGet (eeprom_server_socket, &ah);
	  TCPGet (eeprom_server_socket, &c);
	  count = c;
	  addr = (ah << 8) | al;
	  if (eestate == EE_READADDR)
	    eestate = EE_READING;
	  else
	    {
	      eestate = EE_WRITING;
	      rtc_pause (-1);	/* just to re-read the font */
	    }
	  break;

	case EE_READING:
	  while (count && (space = TCPIsPutReady (eeprom_server_socket)) > 0)
	    {
	      if (space > 128)
		space = 128;
	      if (space > count)
		space = count;
	      eeprom_read (addr, buf, space);
	      addr += space;
	      count -= space;
	      TCPPutArray (eeprom_server_socket, buf, space);
	      TCPFlush (eeprom_server_socket);
	    }
	  if (count == 0)
	    eestate = EE_IDLE;
	  break;

	case EE_WRITING:
	  while (count && (space = TCPIsGetReady (eeprom_server_socket)) > 0)
	    {
	      if (space > count)
		space = count;
	      if ((addr & ~0x7f) != ((addr + space) & ~0x7f))
		space = ((addr + 0x80) & ~0x7f) - addr;

	      TCPGetArray (eeprom_server_socket, buf, space);
	      write_1 (addr, buf, space);
	      addr += space;
	      count -= space;
	    }
	  if (count == 0)
	    {
	      TCPPut (eeprom_server_socket, 0);
	      TCPFlush (eeprom_server_socket);
	      eestate = EE_IDLE;
	    }
	  break;
	}
    }
  else if (TCBStubs[eeprom_server_socket].smState == TCP_CLOSE_WAIT)
    TCPDisconnect (eeprom_server_socket);
}

unsigned int
get_toc (int idx, unsigned int *sizep)
{
  unsigned int toc[2];
  eeprom_read (idx * 4, &toc, 4);
  if (sizep)
    *sizep = toc[1];
  return toc[0];
}
