/* Copyright 2007, 2008 DJ Delorie <dj@delorie.com>
   Distributed under the terms of the GNU General Public License
   version 3, see COPYING-3.0 for details.
*/
#include "pins.h"
#include "funcs.h"
#include "globals.h"
#include "TCPIP Stack/TCPIP.h"

#include "mp3-config.h"

Play_Source play_source;

//static int beep_idx=0, beep_max=sizeof(song);

TCP_SOCKET mp3_url_socket;
static TCP_SOCKET mp3_server_socket;

#define BUFFER_SIZE 4096
// Defined in the linker script, as it is placed right at the end of
// memory, using up some of the "far" space we can't otherwise easily
// use.
extern byte mp3_buffer[BUFFER_SIZE];
static unsigned int buffer_put = 0;
static unsigned int buffer_get = 0;
static unsigned int buffer_count = 0;
static byte playing_silence = 0;
static int last_change_time = 0;

unsigned int beep_start, beep_end, beep_ptr;

/* The Zen timer slowly ramps up the volume from silent to full
   volume.  The range of volumes is 0 (silent) to 255 (full).  */

static long zen_value;
static long zen_scale = 0;
static long zen_start, zen_end;
static uint zen_volume;		/* 0..255 */

/*----------------------------------------------------------------------*/

static void
set_reg (int r, int v)
{
  unsigned char buf[2];
  buf[0] = r;
  buf[1] = v;
  i2c_wr (MP3_I2C, buf, 2);
}

static int
get_reg (int r)
{
  unsigned char buf[1];
  buf[0] = r;
  i2c_wr (MP3_I2C, buf, 1);
  i2c_rd (MP3_I2C, buf, 1);
  return buf[0];
}

/* Some shortcuts */

static void
sta013_soft_reset ()
{
  set_reg (0x10, 1);
}

#if 0
static void
sta013_play (int do_play)
{
  set_reg (0x13, do_play ? 1 : 0);
}

static void
sta013_mute (int do_mute)
{
  set_reg (0x14, do_mute ? 1 : 0);
}
#endif

/* 0..255, 0=off, 255=max */
static void
sta013_volume (int vol)
{
  int atten = ((255 - vol) * 3) / 8;
  set_reg (0x46, atten);
  set_reg (0x48, atten);
}

#if 0
static void
sta013_run (int do_run)
{
  set_reg (0x72, do_run ? 1 : 0);
}

static void
sta013_eq (unsigned int bass_freq, unsigned int treb_freq, int bass_boost,
	   int treb_boost)
{
  int mid_atten;
  set_reg (0x77, treb_freq & 0xff);
  set_reg (0x78, treb_freq >> 8);
  set_reg (0x79, bass_freq & 0xff);
  set_reg (0x7a, bass_freq >> 8);
  bass_boost = CLIP (bass_boost, -12, 12);
  treb_boost = CLIP (treb_boost, -12, 12);
  mid_atten = MAX (0, MAX (treb_boost, bass_boost));
  set_reg (0x7b, treb_boost);
  set_reg (0x7c, bass_boost);
  set_reg (0x7d, mid_atten);
}
#endif

static void
mp3_load_configuration ()
{
  int i;
  for (i = 0; i < sizeof (mp3_config); i += 2)
    set_reg (mp3_config[i], mp3_config[i + 1]);
}

static void
mp3_reset ()
{
  mp3_set_mute (1);
  sta013_soft_reset ();
  mp3_load_configuration ();

#if 0
  {
    unsigned int silence_addr;

    playing_silence = 1;
    buffer_get = 0;
    silence_addr = get_toc (TOC_mp3_silence, &buffer_count);
    eeprom_read (silence_addr, mp3_buffer, buffer_count);
    buffer_put = buffer_count;
  }
#endif
}

static void
mp3_set_source (int mode)
{
  static char *s[] = { "NONE", "BEEP", "URL", "SERVER" };

  if (play_source == mode)
    return;

  last_change_time = eseconds;

  cprintf ("mp3: source %s -> %s\n", s[play_source], s[mode]);

#if 0
  if (play_source != PLAY_NONE)
    mp3_load_configuration ();
#endif

  switch (play_source)
    {
    case PLAY_NONE:
      break;
    case PLAY_BEEP:
      break;
    case PLAY_URL:
      if (mp3_url_socket != INVALID_SOCKET && TCPIsConnected (mp3_url_socket))
	TCPDisconnect (mp3_url_socket);
      break;
    case PLAY_SERVER:
      if (TCPIsConnected (mp3_server_socket))
	TCPDisconnect (mp3_server_socket);
      break;
    }

  switch (mode)
    {
    case PLAY_NONE:
      zen_scale = 0;
      sta013_volume (255);
      mp3_set_mute (1);
      break;
    default:
      mp3_set_mute (0);
      break;
    }

  play_source = mode;
}

void
mp3_init ()
{
  int i;
  for (i = 0; i < 5; i++)
    {
      if (get_reg (1) == 0xac)
	break;
    }
  if (i == 5)
    {
      cput_s ("mp3 chip not present\n");
      return;
    }

  mp3_load_configuration ();

  mp3_server_socket = TCPListen (MP3_SERVER_PORT);
  //cprintf ("mp3 server socket is %d\n", mp3_server_socket);
  mp3_url_socket = INVALID_SOCKET;
  play_source = PLAY_NONE;
  mp3_set_mute (1);
}

void
mp3_process_buffer ()
{
  byte c;

  if (mp3rdy && buffer_count)
    {
      // Note that, while we program the SPI in enhanced (buffered)
      // mode, we don't have to worry about overrunning the buffer
      // here because the SPI is so fast that we just can't feed it
      // fast enough to get ahead.  Thus, we never start a new byte
      // once mp3rdy goes low.
      spi_mp3 ();
      while (mp3rdy && buffer_count)
	{
	  c = mp3_buffer[buffer_get];
	  spi_putc (c);
	  buffer_get = (buffer_get + 1) % BUFFER_SIZE;
	  buffer_count--;
	}
      spi_done ();
      if (buffer_count == 0)
	buffer_put = buffer_get = 0;
    }
}

static void
play_from_socket (TCP_SOCKET s, int reconnect)
{
  int i;
  if (s == INVALID_SOCKET)
    {
      mp3_set_source (PLAY_NONE);
      return;
    }
  if (!TCPIsConnected (s))
    {
      TCPDisconnect (s);
      mp3_set_source (PLAY_NONE);
      if (reconnect)
	{
	  alarm_reconnect ();
	  return;
	}
      return;
    }

  i = TCPIsGetReady (s);
  if (i)
    {
      int ba;
      if (buffer_get > buffer_put)
	ba = buffer_get - buffer_put;
      else if (buffer_count == BUFFER_SIZE)
	return;
      else
	ba = BUFFER_SIZE - buffer_put;
      ba = MIN (ba, i);
      TCPGetArray (s, &mp3_buffer[buffer_put], ba);
      buffer_put = (buffer_put + ba) % BUFFER_SIZE;
      buffer_count += ba;
    }
}

void
mp3_task ()
{

  if (play_source == PLAY_NONE
      && last_change_time && eseconds - (last_change_time + 5) > 0)
    {
      last_change_time = 0;
      mp3_reset ();
    }

#if 0
  {
    static int last_sec = -1;
    if (last_sec != second)
      {
	last_sec = second;
	cprintf ("mp3 %d %d %d %d", play_source, buffer_count, buffer_put,
		 buffer_get);
	switch (play_source)
	  {
	  case PLAY_BEEP:
	    cprintf (" %w %w %w", beep_start, beep_end, beep_ptr);
	    break;
	  case PLAY_SERVER:
	    cprintf (" [%d] %d %d %d %d", mp3_server_socket,
		     TCPIsConnected (mp3_server_socket),
		     TCPIsGetReady (mp3_server_socket),
		     TCPGetRxFIFOFree (mp3_server_socket));
	    break;
	  case PLAY_URL:
	    cprintf (" [%d] %d %d", mp3_url_socket,
		     TCPIsGetReady (mp3_url_socket));
	    break;
	  }
	cprintf ("\n");
      }
  }
#endif

  mp3_process_buffer ();

  mp3_zen_timer_tick ();

  if (playing_silence)
    {
      if (buffer_count)
	return;
      playing_silence = 0;
      mp3_set_mute (status.muted);
    }

  if (play_source != PLAY_SERVER && TCPIsConnected (mp3_server_socket))
    {
      mp3_set_source (PLAY_SERVER);
    }

  switch (play_source)
    {
    case PLAY_NONE:
      break;

    case PLAY_BEEP:
      if (buffer_count < (BUFFER_SIZE - 512))
	{
	  int ba;

	  if (buffer_get > buffer_put)
	    ba = buffer_get - buffer_put;
	  else
	    ba = BUFFER_SIZE - buffer_put;
	  ba = MIN (ba, beep_end - beep_ptr);

	  eeprom_read (beep_ptr, mp3_buffer + buffer_put, ba);
	  buffer_put = (buffer_put + ba) % BUFFER_SIZE;
	  buffer_count += ba;
	  beep_ptr += ba;
	  if (beep_ptr == beep_end)
	    beep_ptr = beep_start;
	}
      break;

    case PLAY_URL:
      play_from_socket (mp3_url_socket, 1);
      break;

    case PLAY_SERVER:
      play_from_socket (mp3_server_socket, 0);
      break;
    }
}

void
mp3_set_mute (int muted)
{
#if 0
  if (status.muted == muted)
    return;
#endif
  status.muted = muted;
  if (muted || playing_silence)
    {
      //cprintf ("Mute: SILENT\n");
      // shut off the amplifiers
      // reconfigure as GPIO output
      AD1PCFGbits.PCFG9 = 1;
      vmute = 0;		// GPIO output, mutes
    }
  else
    {
      //cprintf ("Mute: MUSIC\n");
      // turn on the amplifiers
      // reconfigure as ADC input
      AD1PCFGbits.PCFG9 = 0;
      vmute = 1;		// GPIO input, releases mute
    }
}

void
mp3_play_beep ()
{
  beep_start = get_toc (TOC_default_beep, &beep_end);
  beep_end += beep_start;
  beep_ptr = beep_start;

  mp3_set_source (PLAY_BEEP);
}

void
mp3_stop ()
{
  mp3_set_source (PLAY_NONE);
}

void
mp3_play_url ()
{
  mp3_set_source (PLAY_URL);
}

void
mp3_zen_timer_start (int zen_setting)
{
  if (zen_setting == 0)
    {
      sta013_volume (0xff);
      return;
    }

  /* Up to 225,000 ticks (0x3e418) in the ramp-up.  */
  zen_start = sTick ();
  zen_end = zen_start + zen_setting * TICK_SECOND;

  zen_value = 0;
  zen_scale = 0x7f000000UL / (TICK_SECOND * (unsigned long) zen_setting);
  zen_volume = 255;

  sta013_volume (0);
}

void
mp3_zen_timer_tick ()
{
  static byte ozv = 5;
  byte zv;
  if (!zen_scale)
    return;

  if ((sTick () - zen_end) > 0)
    {
      sta013_volume (zen_volume);
      zen_scale = 0;
      return;
    }
  zen_value = (sTick () - zen_start) * zen_scale;
  zv = (int) (zen_value >> 24) + 128;
  if (zv != ozv)
    {
      ozv = zv;
      sta013_volume (zv * zen_volume / 255);
    }
}

void
mp3_zen_volume (int v)
{
  zen_volume = v / 4 + 192;
  if (zen_volume < 192)
    zen_volume = 192;
  if (zen_volume > 255)
    zen_volume = 255;
  if (!zen_scale)
    sta013_volume (zen_volume);
}
